@@ -10,10 +10,7 @@ use crate::{
     ubig::{Repr::*, UBig},
 };
 use alloc::vec::Vec;
-use core::{
-    borrow::Borrow,
-    convert::{TryFrom, TryInto},
-};
+use core::convert::{TryFrom, TryInto};
 
 impl Default for UBig {
     /// Default value: 0.
@@ -297,69 +294,38 @@ fn round_to_even_adjustment(bits: u32) -> bool {
     bits >= 0b110 || bits == 0b011
 }
 
-/// Implement `impl From<U> for T` using a function.
-macro_rules! impl_from {
-    (impl From<$a:ty> for $b:ty as $f:ident) => {
-        impl From<$a> for $b {
-            fn from(value: $a) -> $b {
-                $f(value)
+macro_rules! ubig_unsigned_conversions {
+    ($t:ty) => {
+        impl From<$t> for UBig {
+            fn from(value: $t) -> UBig {
+                UBig::from_unsigned(value)
             }
         }
-    };
-}
 
-/// Implement `impl TryFrom<U> for T` using a function.
-macro_rules! impl_try_from {
-    (impl TryFrom<$a:ty> for $b:ty as $f:ident) => {
-        impl TryFrom<$a> for $b {
+        impl TryFrom<UBig> for $t {
             type Error = OutOfBoundsError;
 
-            fn try_from(value: $a) -> Result<$b, OutOfBoundsError> {
-                $f(value)
+            fn try_from(value: UBig) -> Result<$t, OutOfBoundsError> {
+                value.try_to_unsigned()
+            }
+        }
+
+        impl TryFrom<&UBig> for $t {
+            type Error = OutOfBoundsError;
+
+            fn try_from(value: &UBig) -> Result<$t, OutOfBoundsError> {
+                value.try_to_unsigned()
             }
         }
     };
 }
 
-impl_from!(impl From<u8> for UBig as ubig_from_unsigned);
-impl_from!(impl From<u16> for UBig as ubig_from_unsigned);
-impl_from!(impl From<u32> for UBig as ubig_from_unsigned);
-impl_from!(impl From<u64> for UBig as ubig_from_unsigned);
-impl_from!(impl From<u128> for UBig as ubig_from_unsigned);
-impl_from!(impl From<usize> for UBig as ubig_from_unsigned);
-
-impl_try_from!(impl TryFrom<i8> for UBig as ubig_from_signed);
-impl_try_from!(impl TryFrom<i16> for UBig as ubig_from_signed);
-impl_try_from!(impl TryFrom<i32> for UBig as ubig_from_signed);
-impl_try_from!(impl TryFrom<i64> for UBig as ubig_from_signed);
-impl_try_from!(impl TryFrom<i128> for UBig as ubig_from_signed);
-impl_try_from!(impl TryFrom<isize> for UBig as ubig_from_signed);
-
-impl_try_from!(impl TryFrom<UBig> for u8 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for u16 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for u32 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for u64 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for u128 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for usize as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for u8 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for u16 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for u32 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for u64 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for u128 as unsigned_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for usize as unsigned_from_ubig);
-
-impl_try_from!(impl TryFrom<UBig> for i8 as signed_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for i16 as signed_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for i32 as signed_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for i64 as signed_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for i128 as signed_from_ubig);
-impl_try_from!(impl TryFrom<UBig> for isize as signed_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for i8 as signed_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for i16 as signed_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for i32 as signed_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for i64 as signed_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for i128 as signed_from_ubig);
-impl_try_from!(impl TryFrom<&UBig> for isize as signed_from_ubig);
+ubig_unsigned_conversions!(u8);
+ubig_unsigned_conversions!(u16);
+ubig_unsigned_conversions!(u32);
+ubig_unsigned_conversions!(u64);
+ubig_unsigned_conversions!(u128);
+ubig_unsigned_conversions!(usize);
 
 impl From<bool> for UBig {
     fn from(b: bool) -> UBig {
@@ -367,45 +333,73 @@ impl From<bool> for UBig {
     }
 }
 
-impl_from!(impl From<u8> for IBig as ibig_from_unsigned);
-impl_from!(impl From<u16> for IBig as ibig_from_unsigned);
-impl_from!(impl From<u32> for IBig as ibig_from_unsigned);
-impl_from!(impl From<u64> for IBig as ibig_from_unsigned);
-impl_from!(impl From<u128> for IBig as ibig_from_unsigned);
-impl_from!(impl From<usize> for IBig as ibig_from_unsigned);
+macro_rules! ubig_signed_conversions {
+    ($t:ty) => {
+        impl TryFrom<$t> for UBig {
+            type Error = OutOfBoundsError;
 
-impl_from!(impl From<i8> for IBig as ibig_from_signed);
-impl_from!(impl From<i16> for IBig as ibig_from_signed);
-impl_from!(impl From<i32> for IBig as ibig_from_signed);
-impl_from!(impl From<i64> for IBig as ibig_from_signed);
-impl_from!(impl From<i128> for IBig as ibig_from_signed);
-impl_from!(impl From<isize> for IBig as ibig_from_signed);
+            fn try_from(value: $t) -> Result<UBig, OutOfBoundsError> {
+                UBig::try_from_signed(value)
+            }
+        }
 
-impl_try_from!(impl TryFrom<IBig> for u8 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for u16 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for u32 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for u64 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for u128 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for usize as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for u8 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for u16 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for u32 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for u64 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for u128 as unsigned_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for usize as unsigned_from_ibig);
+        impl TryFrom<UBig> for $t {
+            type Error = OutOfBoundsError;
 
-impl_try_from!(impl TryFrom<IBig> for i8 as signed_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for i16 as signed_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for i32 as signed_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for i64 as signed_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for i128 as signed_from_ibig);
-impl_try_from!(impl TryFrom<IBig> for isize as signed_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for i8 as signed_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for i16 as signed_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for i32 as signed_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for i64 as signed_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for i128 as signed_from_ibig);
-impl_try_from!(impl TryFrom<&IBig> for isize as signed_from_ibig);
+            fn try_from(value: UBig) -> Result<$t, OutOfBoundsError> {
+                value.try_to_signed()
+            }
+        }
+
+        impl TryFrom<&UBig> for $t {
+            type Error = OutOfBoundsError;
+
+            fn try_from(value: &UBig) -> Result<$t, OutOfBoundsError> {
+                value.try_to_signed()
+            }
+        }
+    };
+}
+
+ubig_signed_conversions!(i8);
+ubig_signed_conversions!(i16);
+ubig_signed_conversions!(i32);
+ubig_signed_conversions!(i64);
+ubig_signed_conversions!(i128);
+ubig_signed_conversions!(isize);
+
+macro_rules! ibig_unsigned_conversions {
+    ($t:ty) => {
+        impl From<$t> for IBig {
+            fn from(value: $t) -> IBig {
+                IBig::from_unsigned(value)
+            }
+        }
+
+        impl TryFrom<IBig> for $t {
+            type Error = OutOfBoundsError;
+
+            fn try_from(value: IBig) -> Result<$t, OutOfBoundsError> {
+                value.try_to_unsigned()
+            }
+        }
+
+        impl TryFrom<&IBig> for $t {
+            type Error = OutOfBoundsError;
+
+            fn try_from(value: &IBig) -> Result<$t, OutOfBoundsError> {
+                value.try_to_unsigned()
+            }
+        }
+    };
+}
+
+ibig_unsigned_conversions!(u8);
+ibig_unsigned_conversions!(u16);
+ibig_unsigned_conversions!(u32);
+ibig_unsigned_conversions!(u64);
+ibig_unsigned_conversions!(u128);
+ibig_unsigned_conversions!(usize);
 
 impl From<bool> for IBig {
     fn from(b: bool) -> IBig {
@@ -413,6 +407,39 @@ impl From<bool> for IBig {
     }
 }
 
+macro_rules! ibig_signed_conversions {
+    ($t:ty) => {
+        impl From<$t> for IBig {
+            fn from(value: $t) -> IBig {
+                IBig::from_signed(value)
+            }
+        }
+
+        impl TryFrom<IBig> for $t {
+            type Error = OutOfBoundsError;
+
+            fn try_from(value: IBig) -> Result<$t, OutOfBoundsError> {
+                value.try_to_signed()
+            }
+        }
+
+        impl TryFrom<&IBig> for $t {
+            type Error = OutOfBoundsError;
+
+            fn try_from(value: &IBig) -> Result<$t, OutOfBoundsError> {
+                value.try_to_signed()
+            }
+        }
+    };
+}
+
+ibig_signed_conversions!(i8);
+ibig_signed_conversions!(i16);
+ibig_signed_conversions!(i32);
+ibig_signed_conversions!(i64);
+ibig_signed_conversions!(i128);
+ibig_signed_conversions!(isize);
+
 impl From<UBig> for IBig {
     fn from(x: UBig) -> IBig {
         IBig::from_sign_magnitude(Positive, x)
@@ -447,43 +474,65 @@ impl TryFrom<&IBig> for UBig {
     }
 }
 
-/// Convert an unsigned primitive to [UBig].
-fn ubig_from_unsigned<T>(x: T) -> UBig
-where
-    T: PrimitiveUnsigned,
-{
-    match x.try_into() {
-        Ok(w) => UBig::from_word(w),
-        Err(_) => {
-            let repr = x.to_le_bytes();
-            UBig::from_le_bytes(repr.as_ref())
+impl UBig {
+    /// Convert an unsigned primitive to [UBig].
+    pub(crate) fn from_unsigned<T>(x: T) -> UBig
+    where
+        T: PrimitiveUnsigned,
+    {
+        match x.try_into() {
+            Ok(w) => UBig::from_word(w),
+            Err(_) => {
+                let repr = x.to_le_bytes();
+                UBig::from_le_bytes(repr.as_ref())
+            }
         }
     }
-}
 
-/// Try to convert a signed primitive to [UBig].
-fn ubig_from_signed<T>(x: T) -> Result<UBig, OutOfBoundsError>
-where
-    T: PrimitiveSigned,
-{
-    match T::Unsigned::try_from(x) {
-        Ok(u) => Ok(ubig_from_unsigned(u)),
-        Err(_) => Err(OutOfBoundsError),
-    }
-}
-
-/// Try to convert [UBig] to an unsigned primitive.
-fn unsigned_from_ubig<T, B>(num: B) -> Result<T, OutOfBoundsError>
-where
-    T: PrimitiveUnsigned,
-    B: Borrow<UBig>,
-{
-    match num.borrow().repr() {
-        Small(w) => match T::try_from(*w) {
-            Ok(val) => Ok(val),
+    /// Try to convert a signed primitive to [UBig].
+    fn try_from_signed<T>(x: T) -> Result<UBig, OutOfBoundsError>
+    where
+        T: PrimitiveSigned,
+    {
+        match T::Unsigned::try_from(x) {
+            Ok(u) => Ok(UBig::from_unsigned(u)),
             Err(_) => Err(OutOfBoundsError),
-        },
-        Large(buffer) => unsigned_from_words(buffer),
+        }
+    }
+
+    /// Try to convert [UBig] to an unsigned primitive.
+    pub(crate) fn try_to_unsigned<T>(&self) -> Result<T, OutOfBoundsError>
+    where
+        T: PrimitiveUnsigned,
+    {
+        match self.repr() {
+            Small(w) => match T::try_from(*w) {
+                Ok(val) => Ok(val),
+                Err(_) => Err(OutOfBoundsError),
+            },
+            Large(buffer) => unsigned_from_words(buffer),
+        }
+    }
+
+    /// Try to convert [UBig] to a signed primitive.
+    fn try_to_signed<T>(&self) -> Result<T, OutOfBoundsError>
+    where
+        T: PrimitiveSigned,
+    {
+        match self.repr() {
+            Small(w) => T::try_from(*w).map_err(|_| OutOfBoundsError),
+            Large(buffer) => {
+                let u: T::Unsigned = unsigned_from_words(buffer)?;
+                u.try_into().map_err(|_| OutOfBoundsError)
+            }
+        }
+    }
+
+    pub(crate) fn from_ibig_panic_on_overflow(x: IBig) -> UBig {
+        match UBig::try_from(x) {
+            Ok(v) => v,
+            Err(_) => panic!("UBig overflow"),
+        }
     }
 }
 
@@ -511,58 +560,29 @@ where
     }
 }
 
-/// Try to convert [UBig] to a signed primitive.
-fn signed_from_ubig<T, B>(num: B) -> Result<T, OutOfBoundsError>
-where
-    T: PrimitiveSigned,
-    B: Borrow<UBig>,
-{
-    match num.borrow().repr() {
-        Small(w) => T::try_from(*w).map_err(|_| OutOfBoundsError),
-        Large(buffer) => {
-            let u: T::Unsigned = unsigned_from_words(buffer)?;
-            u.try_into().map_err(|_| OutOfBoundsError)
+impl IBig {
+    /// Convert an unsigned primitive to [IBig].
+    pub(crate) fn from_unsigned<T: PrimitiveUnsigned>(x: T) -> IBig {
+        IBig::from(UBig::from_unsigned(x))
+    }
+
+    /// Convert a signed primitive to [IBig].
+    pub(crate) fn from_signed<T: PrimitiveSigned>(x: T) -> IBig {
+        let (sign, mag) = x.to_sign_magnitude();
+        IBig::from_sign_magnitude(sign, UBig::from_unsigned(mag))
+    }
+
+    /// Try to convert [IBig] to an unsigned primitive.
+    pub(crate) fn try_to_unsigned<T: PrimitiveUnsigned>(&self) -> Result<T, OutOfBoundsError> {
+        match self.sign() {
+            Positive => self.magnitude().try_to_unsigned(),
+            Negative => Err(OutOfBoundsError),
         }
     }
-}
 
-/// Convert an unsigned primitive to [IBig].
-fn ibig_from_unsigned<T>(x: T) -> IBig
-where
-    T: PrimitiveUnsigned,
-{
-    IBig::from(ubig_from_unsigned(x))
-}
-
-/// Convert a signed primitive to [IBig].
-fn ibig_from_signed<T>(x: T) -> IBig
-where
-    T: PrimitiveSigned,
-{
-    let (sign, mag) = x.to_sign_magnitude();
-    IBig::from_sign_magnitude(sign, ubig_from_unsigned(mag))
-}
-
-/// Try to convert [IBig] to an unsigned primitive.
-fn unsigned_from_ibig<T, B>(num: B) -> Result<T, OutOfBoundsError>
-where
-    T: PrimitiveUnsigned,
-    B: Borrow<IBig>,
-{
-    let num = num.borrow();
-    match num.sign() {
-        Positive => unsigned_from_ubig(num.magnitude()),
-        Negative => Err(OutOfBoundsError),
+    /// Try to convert [IBig] to an signed primitive.
+    pub(crate) fn try_to_signed<T: PrimitiveSigned>(&self) -> Result<T, OutOfBoundsError> {
+        let u: T::Unsigned = self.magnitude().try_to_unsigned()?;
+        T::try_from_sign_magnitude(self.sign(), u)
     }
 }
-
-/// Try to convert [IBig] to an signed primitive.
-fn signed_from_ibig<T, B>(num: B) -> Result<T, OutOfBoundsError>
-where
-    T: PrimitiveSigned,
-    B: Borrow<IBig>,
-{
-    let num = num.borrow();
-    let u: T::Unsigned = unsigned_from_ubig(num.magnitude())?;
-    T::try_from_sign_magnitude(num.sign(), u)
-}
